from marshmallow import Schema, EXCLUDE, fields

class ExecutableJobConfigSchema(Schema):
    queue = fields.String(
        required=True,
        description="Name of queue on which to place task.",
        example="my-default-queue",
    )
    task = fields.String(
        required=True,
        description="Path to task to invoke.",
        example="my_app.module.method",
    )

class ExecutableJobSchema(Schema):
    """Definition of a single schedule entry"""
    class Meta:
        unknown = EXCLUDE

    name = fields.String(
        required=True,
        description="Name of schedule entry.",
        example="My Scheduled Task",
    )
    schemaVersion = fields.Integer(required=True)
    config = fields.Dict(required=True)
    enabled = fields.Boolean(
        required=True, description="Whether entry is enabled.", example=True
    )
    config = fields.Nested(
        ExecutableJobConfigSchema,
        required=True,
        description="Configuration information for this job.",
    )




def get_clean_validation_messages(validation_error):
    """
    Extract and format clean validation error messages.

    Args:
        validation_error (ValidationError): The Marshmallow ValidationError instance.

    Returns:
        str: A formatted string with all validation error messages.
    """

    def format_errors(errors, parent_key=""):
        messages = []
        for key, value in errors.items():
            full_key = f"{parent_key}.{key}" if parent_key else key
            if isinstance(value, dict):
                # Recursively format nested errors
                messages.extend(format_errors(value, full_key))
            else:
                # Append error messages
                messages.append(f"{full_key}: {', '.join(value)}")
        return messages

    return "\n".join(format_errors(validation_error.messages))



""" Definition of the `pypeline.yaml` file.

If using, a basic file may look like::
    serviceConfig:
        - name: pypeline-worker
            registeredTasks:
                - handler: pypeline_demo_client.workers.demo_worker.demo_worker_task
                - handler: pypeline_demo_client.workers.demo_worker.demo_model_task

    pipelines:
        demo-pipeline:
            name: demo-pipeline
            description: Demo Pipeline.
            schemaVersion: 1
            config:
                dagAdjacency:
                    node_a:
                        - node_b
                        - node_c
                metadata:
                    maxRetry: 3
                    maxTtl: 60
                    queue: default-task-queue
                taskDefinitions:
                    node_a:
                        handler: pypeline_demo_client.workers.demo_pipeline.demo_pipeline_node_a
                    node_b:
                        handler: pypeline_demo_client.workers.demo_pipeline.demo_pipeline_node_b
                        queue: node-b-queue
                    node_c:
                        handler: pypeline_demo_client.workers.demo_pipeline.demo_pipeline_node_c

    scheduledTasks:
        demo-model-task:
            name: Demo Model Task
            enabled: true
            config:
                task: pypeline_demo_client.workers.demo_worker.demo_model_task
                queue: default-task-queue
                schedule:
                    minute: '*'
                    hour: '*'
                    dayOfWeek: '*'
                    dayOfMonth: '*'
                    monthOfYear: '*'
            schemaVersion: 1

"""
import re
import os
import logging
import pkg_resources
import yaml
from yaml.loader import SafeLoader
from marshmallow import Schema, fields, pre_load, EXCLUDE, INCLUDE, validates_schema
from marshmallow.exceptions import ValidationError

from pypeline.executable_job_config_schema import ExecutableJobSchema
from pypeline.utils.module_utils import PypelineModuleLoader, normalized_pkg_name
from pypeline.constants import PYPELINE_YAML_PATH, PYPELINE_CLIENT_PKG_NAME
from pypeline.pipeline_config_schema import BasePipelineSchema
from pypeline.schedule_config_schema import BaseScheduleSchema

logger = logging.getLogger(__name__)


class InvalidPackagePath(Exception):
    pass


class InvalidPypelineConfig(Exception):
    pass


class MissingPypelineConfig(Exception):
    pass


class ExcludeUnknownSchema(Schema):
    class Meta:
        unknown = EXCLUDE


class NameSchema(Schema):
    """Validated name string field."""

    name = fields.String(
        required=True,
        description="Name for service or image. Must include "
        "only alphanumeric characters along with `_` and `-`.",
        example="my-service-name",
    )

    @pre_load
    def validate_characters(self, item, **kwargs):
        """Ensure name field conforms to allowed characters"""
        valid_chars = r"^[\w\d\-\_]+$"
        if not bool(re.match(valid_chars, item["name"])):
            raise ValueError(
                f"Invalid name: {item['name']}. Only alphanumeric characters "
                "allowed along with `-` and `_`."
            )
        return item


class PypelineRegisteredTaskDetailConfigSchema(Schema):
    handler = fields.String(
        required=True,
        description="Full path to the Method handles work / pipeline tasks.",
        example="pypeline_customer_client.workers.worker_group.useful_worker",
    )

    event = fields.Raw(
        required=False,
        unknown=INCLUDE,
        description="Arbitrary user data, passed through `event` arg in task.",
    )


class PypelineCeleryWorkerConfigSchema(Schema):
    """Attributes for a celery worker.  This worker will run all of the
    pipelines and scheduled tasks.
    """

    registeredTasks = fields.List(
        fields.Nested(PypelineRegisteredTaskDetailConfigSchema, required=True),
        required=False,
        _required=True,
        description="List of task handlers to register for to your Pypeline app.",
    )


class PypelineServiceConfigSchema(
    ExcludeUnknownSchema, PypelineCeleryWorkerConfigSchema, NameSchema
):
    """Base service config object definition for workers."""

    pass


class PypelineYamlSchema(ExcludeUnknownSchema):
    """The primary `pypeline.yaml` file schema. This defines all available
    properties in a valid Pypeline configuration file.
    """

    serviceConfig = fields.List(
        fields.Nested(
            PypelineServiceConfigSchema,
            required=True,
            description="Core service configuration.",
        ),
        description="List of workers for Pypeline to manage.",
        required=True,
    )

    pipelines = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(BasePipelineSchema),
        description="List of pipelines",
        required=False,
    )

    scheduledTasks = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(BaseScheduleSchema),
        description="List of scheduled tasks",
        required=False,
    )

    executableJobs = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(ExecutableJobSchema),
        description="List of executable jobs",
        required=False,
        allow_none=True,
    )

    def validate_errors(self, schema: Schema, value: dict):
        """Run Marshmallow validate() and raise if any errors"""
        schema = schema()
        errors = schema.validate(value)
        if len(errors.keys()) > 0:
            raise ValidationError(errors)

    @validates_schema
    def validate_schema(self, data, **kwargs):
        """Additional validation.

        Nested fields that are not required are not validated by Marshmallow
        by default. Do a single level down of validation for now.

        imageConfig can provide *either* an install command for Pypeline
        to use to build the image for customer *or* a Docker repository
        for Pypeline to pull.
        """
        # Vaidate nested
        key_schema_pairs = (("serviceConfig", PypelineServiceConfigSchema),)
        for k_s in key_schema_pairs:
            val = data.get(k_s[0], None)
            if val is not None:
                if type(val) == list:
                    for v in val:
                        self.validate_errors(k_s[1], v)
                else:
                    self.validate_errors(k_s[1], val)

        # Validate the services. We list every service schema field as not
        # required in order to use them as mixins for a generic service object,
        # however, they ARE required, so validate here using the custom
        # metadata property `_required`. Default to value of `required`.
        for service in data.get("serviceConfig"):
            schema = PypelineCeleryWorkerConfigSchema
            for field in schema().fields:
                try:
                    if (
                        schema()
                        .fields[field]
                        .metadata.get(
                            "_required", getattr(schema().fields[field], "required")
                        )
                    ):
                        assert field in service
                except AssertionError:
                    raise ValidationError(f"`{field}` missing in worker definition.")

        # Validate unique pipeline ids
        if "pipelines" in data:
            pipeline_ids = set()
            for pipeline_id, pipeline_data in data["pipelines"].items():
                if pipeline_id in pipeline_ids:
                    raise ValidationError("All pipeline ids must be unique!")
                pipeline_ids.add(pipeline_id)
                schema_version = pipeline_data["schemaVersion"]
                PipelineSchema = BasePipelineSchema.get_by_version(schema_version)
                self.validate_errors(PipelineSchema, pipeline_data)

        # Validate unique scheduled tasks names
        if "scheduledTasks" in data:
            task_ids = set()
            for task_id, task_data in data["scheduledTasks"].items():
                if task_id in task_ids:
                    raise ValidationError("All schedule ids must be unique!")
                task_ids.add(task_id)
                schema_version = task_data["schemaVersion"]
                TaskSchema = BaseScheduleSchema.get_by_version(schema_version)
                self.validate_errors(TaskSchema, task_data)


class YamlPatternConstructor:
    """Adds a pattern resolver + constructor to PyYaml.

    Typical/deault usage is for parsing environment variables
    in a yaml file but this can be used for any pattern you provide.

    See: https://pyyaml.org/wiki/PyYAMLDocumentation
    """

    def __init__(self, env_var_pattern: str = None, add_constructor: bool = True):
        self.env_var_pattern = env_var_pattern
        if self.env_var_pattern is None:
            # Default pattern is: ${VAR:default}
            self.env_var_pattern = r"^\$\{(.*)\}$"
        self.path_matcher = re.compile(self.env_var_pattern)

        if add_constructor:
            self.add_constructor()

    def _path_constructor(self, loader, node):
        """Extract the matched value, expand env variable,
        and replace the match

        TODO: Would need to update this (specifically the parsing) if any
        pattern other than our default (or a highly compatible variation)
        is provided.
        """
        # Try to match the correct env variable pattern in this node's value
        # If the value does not match the pattern, return None (which means
        # this node will not be parsed for ENV variables and instead just
        # returned as-is).
        env_var_name = re.match(self.env_var_pattern, node.value)
        try:
            env_var_name = env_var_name.group(1)
        except AttributeError:
            return None

        # If we get down here, then the 'node.value' matches our specified
        # pattern, so try to parse. env_var_name is the value inside ${...}.
        # Split on `:`, which is our delimiter for default values.
        env_var_name_split = env_var_name.split(":")

        # Attempt to retrieve the environment variable...from the environment
        env_var = os.environ.get(env_var_name_split[0], None)

        if env_var is None:  # Nothing found in environment
            # If a default was provided (e.g. VAR:default), return that.
            # We join anything after first element because the default
            # value might be a URL or something with a colon in it
            # which would have 'split' above
            if len(env_var_name_split) > 1:
                return ":".join(env_var_name_split[1:])
            return "unset"  # Return 'unset' if not in environ nor default
        return env_var

    def add_constructor(self):
        """Initialize PyYaml with ability to resolve/load environment
        variables defined in a yaml template when they exist in
        the environment.

        Add to SafeLoader in addition to standard Loader.
        """
        # Add the `!env_var` tag to any scalar (value) that matches the
        # pattern self.path_matcher. This allows the template to be much more
        # intuitive vs needing to add !env_var to the beginning of each value
        yaml.add_implicit_resolver("!env_var", self.path_matcher)
        yaml.add_implicit_resolver("!env_var", self.path_matcher, Loader=SafeLoader)

        # Add constructor for the tag `!env_var`, which is a function that
        # converts a node of a YAML representation graph to a native Python
        # object.
        yaml.add_constructor("!env_var", self._path_constructor)
        yaml.add_constructor("!env_var", self._path_constructor, Loader=SafeLoader)


def parse_config_file(pypeline_yaml: str):
    """Parse the `pypeline.yaml` file when it's been loaded.

    Arguments:
        pypeline_yaml (required): String of loaded pypeline.yaml file.
    """
    YamlPatternConstructor()  # Add our env variable parser
    try:
        pypeline_yaml_schema = PypelineYamlSchema()
        # First suss out yaml issues
        pypeline_config = yaml.safe_load(pypeline_yaml)
        # Then schema issues
        pypeline_config = pypeline_yaml_schema.load(pypeline_config)
    except ValidationError as e:
        msg = "Invalid Pypeline configuration due to {}".format(e.messages)
        logger.error(msg)
        raise InvalidPypelineConfig(msg)
    except Exception as e:
        msg = (
            "Invalid Pypeline configuration, likely due to invalid "
            "YAML formatting ..."
        )
        logger.exception("{} {}".format(msg, e))
        raise InvalidPypelineConfig(msg)
    return pypeline_config


def _get_pkg_name(pkg_name: str) -> str:
    """Retrieve the normalized package name."""
    if pkg_name is None:
        pkg_name = PYPELINE_CLIENT_PKG_NAME  # From environment
        if pkg_name is None:
            return None
    return normalized_pkg_name(pkg_name)


def load_pypeline_config(
    pkg_name: str = None, pypeline_yaml_filename: str = None, as_dict: bool = True
):
    """Load and parse the `pypeline.yaml` file. Issue usable exceptions for
    known error modes so bootstrapping can handle appropriately.

    Arguments:
        pkg_name (required): Directory name for your Python
            package. e.g. my_package_name . If none provided, will check
            environment for `PYPELINE_CLIENT_PKG_NAME`. If not found,
            will exit.
        pypeline_yaml_filename (optional): Relative path to find your
            `pypeline.yaml` configuration file. Defaults to `pypeline.yaml`
            which should be found inside your `pkg_name`
        as_dict (optional): If true (default), return the loaded pypeline
            configuration as a dictionary. If false, return the loaded
            string value of the yaml file.
    """
    if pypeline_yaml_filename is None:
        pypeline_yaml_filename = PYPELINE_YAML_PATH

    logger.info(
        f"Loading `pypeline.yaml` from package `{pkg_name}` "
        f"and file location `{pypeline_yaml_filename}` ..."
    )
    pypeline_config = None

    pkg_name = _get_pkg_name(pkg_name)

    if pkg_name is None:  # Nothing to retrieve at this point
        logger.warning("Unable to retrieve pypeline.yaml configuration ...")
        return pypeline_config

    try:
        pypeline_config_path = pkg_resources.resource_filename(
            pkg_name, pypeline_yaml_filename
        )
    except Exception as e:
        msg = (
            "Either pkg_name ({}) or pypeline_yaml_filename ({}) is "
            "invalid ...".format(pkg_name, pypeline_yaml_filename)
        )
        logger.error("{} ... {}".format(msg, e))
        raise InvalidPackagePath(e)

    try:
        with open(pypeline_config_path, "r") as f:
            pypeline_yaml = f.read()
            pypeline_config = parse_config_file(pypeline_yaml)
    except InvalidPypelineConfig as e:
        raise
    except FileNotFoundError as e:
        msg = "Pypeline config file could not be found at path {} ...".format(
            pypeline_config_path
        )
        raise MissingPypelineConfig(msg)
    except Exception as e:
        raise e
    if as_dict:
        return pypeline_config
    return yaml.safe_dump(pypeline_config)


def load_client_config_and_version(
    pkg_name: str = None, pypeline_yaml_filename: str = None
):
    """Load and parse the `pypeline.yaml` file and a client package's version.

        Arguments:
            pkg_name (required): Directory name for your Python
                package. e.g. my_package_name . If none provided, will check
                environment for `PYPELINE_CLIENT_PKG_NAME`. If not found,
                will exit.
            pypeline_yaml_filename (optional): Relative path to find your
                `pypeline.yaml` configuration file. Defaults to `pypeline.yaml`
                which should be found inside your `pkg_name`
            as_dict (optional): If true (default), return the loaded pypeline
                configuration as a dictionary. If false, return the loaded
                string value of the yaml file.

    For this to work properly, the provided package must be installed in the
    same environment as this Pypeline package and it must have a `__version__`
    variable inside its `__init__.py` file, e.g. `__version__ = '0.0.0'`
    """
    pypeline_config = None
    client_version = None

    pkg_name = _get_pkg_name(pkg_name)

    try:
        loader = PypelineModuleLoader()
        pkg = loader.get_module(pkg_name + ".__init__")
        client_version = getattr(pkg, "__version__", "0.0.0")
        pypeline_config = load_pypeline_config(pkg_name, pypeline_yaml_filename)
    except MissingPypelineConfig as e:
        logger.error(e)
    except InvalidPypelineConfig as e:
        logger.error(e)
    except InvalidPackagePath as e:
        logger.error(e)
    except Exception as e:
        logger.error(
            "Unable to load client's pkg __version__ or "
            "{} config file for package: {} ... {}".format(
                pypeline_yaml_filename, pkg_name, e
            )
        )

    return pypeline_config, client_version



from copy import copy

import networkx as nx
from dramatiq import Middleware

from pypeline.barrier import LockingParallelBarrier
from pypeline.constants import PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
from pypeline.utils.module_utils import get_callable
from pypeline.utils.pipeline_utils import get_execution_graph
from pypeline.utils.dramatiq_utils import register_lazy_actor


class PypelineMiddleware(Middleware):

    def __init__(self, redis_url):
        self.redis_url = redis_url

    def after_process_message(self, broker, message, *, result=None, exception=None):

        if exception is not None:
            return

        if "pipeline" not in message.options:
            return

        pipeline = message.options["pipeline"]
        max_retries = message.options.get("max_retries", None)
        pipeline_config = pipeline["config"]
        task_replacements = message.options["task_replacements"]
        execution_id = message.options["execution_id"]
        task_definitions = pipeline_config["taskDefinitions"]
        task_name = message.options["task_name"]
        task_key = f"{execution_id}-{task_name}"

        # Signal to other jobs that current task is finished
        locking_parallel_barrier = LockingParallelBarrier(
            self.redis_url,
            task_key=task_key,
            lock_key=f"{message.options['base_case_execution_id']}-lock",
        )
        try:
            locking_parallel_barrier.acquire_lock(timeout=10)
            _ = locking_parallel_barrier.decrement_task_count()
        finally:
            locking_parallel_barrier.release_lock()

        graph = get_execution_graph(pipeline_config)
        children_tasks = pipeline_config["dagAdjacency"].get(task_name, [])

        messages = []
        for child in children_tasks:
            child_ancestors = sorted(graph.predecessors(child))

            ancestor_tasks_complete = True

            for ancestor in child_ancestors:
                ancestor_task_key = f"{execution_id}-{ancestor}"

                locking_parallel_barrier = LockingParallelBarrier(
                    self.redis_url,
                    task_key=ancestor_task_key,
                    lock_key=f"{message.options['base_case_execution_id']}-lock",
                )
                try:
                    locking_parallel_barrier.acquire_lock(
                        timeout=PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
                    )

                    if locking_parallel_barrier.task_exists():
                        remaining_tasks = locking_parallel_barrier.get_task_count()
                    else:
                        remaining_tasks = None
                finally:
                    locking_parallel_barrier.release_lock()

                # If the lock didn't exist for the current tasks execution id then it would indicate
                # that this is the start of a new scenario.  Therefore we need to find the ancestor
                # that is executed in the base case execution id and make sure it has completed
                if remaining_tasks is None:
                    ancestor_task_key = (
                        f"{message.options['base_case_execution_id']}-{ancestor}"
                    )

                    locking_parallel_barrier = LockingParallelBarrier(
                        self.redis_url,
                        task_key=ancestor_task_key,
                        lock_key=f"{message.options['base_case_execution_id']}-lock",
                    )
                    try:
                        locking_parallel_barrier.acquire_lock(
                            timeout=PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL
                        )

                        if locking_parallel_barrier.task_exists():
                            remaining_tasks = locking_parallel_barrier.get_task_count()
                    finally:
                        locking_parallel_barrier.release_lock()
                if remaining_tasks is None or remaining_tasks >= 1:
                    ancestor_tasks_complete = False
                    break

            # If the child's ancestor tasks aren't complete move onto the next child to check
            if not ancestor_tasks_complete:
                break

            # Handle situation where base case kicks off new scenario
            if (
                message.options["base_case_execution_id"]
                == message.options["execution_id"]
            ):
                for scenario in message.options["scenarios"]:
                    child_predecessors = list(graph.predecessors(child))
                    if (
                        child in scenario["tasksToRunInScenario"]
                        and task_name in child_predecessors
                        and task_name not in scenario["tasksToRunInScenario"]
                    ):
                        task_key = f"{scenario['execution_id']}-{child}"
                        locking_parallel_barrier = LockingParallelBarrier(
                            self.redis_url,
                            task_key=task_key,
                            lock_key=f"{message.options['base_case_execution_id']}-lock",
                        )
                        locking_parallel_barrier.set_task_count(1)
                        handler = task_definitions[child]["handlers"][
                            task_replacements.get(child, 0)
                        ]
                        server_type = task_definitions[child].get("serverType", None)

                        lazy_actor = register_lazy_actor(
                            broker,
                            get_callable(handler),
                            pipeline_config["metadata"],
                            server_type,
                        )
                        scenario_message = lazy_actor.message()
                        scenario_message.options["pipeline"] = pipeline
                        if max_retries is not None:
                            scenario_message.options["max_retries"] = max_retries
                        scenario_message.options["task_replacements"] = (
                            task_replacements
                        )
                        scenario_message.options["execution_id"] = scenario[
                            "execution_id"
                        ]

                        scenario_message.options["task_name"] = child
                        scenario_message.options["base_case_execution_id"] = (
                            message.options["base_case_execution_id"]
                        )
                        scenario_message.options["scenarios"] = message.options[
                            "scenarios"
                        ]
                        if "settings" in message.kwargs:
                            scenario_message.kwargs["settings"] = copy(
                                message.kwargs["settings"]
                            )
                            scenario_message.kwargs["settings"]["execution_id"] = (
                                scenario["execution_id"]
                            )
                        messages.append(scenario_message)

            # Kick off child task for current scenario
            task_key = f"{execution_id}-{child}"
            locking_parallel_barrier = LockingParallelBarrier(
                self.redis_url,
                task_key=task_key,
                lock_key=f"{message.options['base_case_execution_id']}-lock",
            )
            locking_parallel_barrier.set_task_count(1)
            handler = task_definitions[child]["handlers"][
                task_replacements.get(child, 0)
            ]
            server_type = task_definitions[child].get("serverType", None)
            lazy_actor = register_lazy_actor(
                broker,
                get_callable(handler),
                pipeline_config["metadata"],
                server_type,
            )

            child_message = lazy_actor.message()
            child_message.options["pipeline"] = pipeline
            if max_retries is not None:
                child_message.options["max_retries"] = max_retries
            child_message.options["task_replacements"] = task_replacements
            child_message.options["execution_id"] = execution_id
            child_message.options["task_name"] = child
            child_message.options["base_case_execution_id"] = message.options[
                "base_case_execution_id"
            ]
            child_message.options["scenarios"] = message.options["scenarios"]
            if "settings" in message.kwargs:
                child_message.kwargs["settings"] = message.kwargs["settings"]
                child_message.kwargs["settings"]["execution_id"] = message.options[
                    "execution_id"
                ]

            messages.append(child_message)

        for new_message in messages:
            broker.enqueue(new_message)



""" Sermos' Flask Implementation and Tooling. Convenience imports here.
"""
import logging

logger = logging.getLogger(__name__)

try:
    from flask_smorest import Blueprint, Api
    from flask import abort
except Exception as e:
    logger.error("Unable to import Web services (Blueprint, API, abort)" f" ... {e}")

try:
    from pypeline.flask.flask_pypeline import FlaskPypeline
except Exception as e:
    logger.exception("Unable to import Sermos services (FlaskPypeline)" f" ... {e}")



from __future__ import annotations

import copy
import json
import time
import typing
from uuid import uuid4

from dramatiq.broker import get_broker
from dramatiq.results import ResultMissing
from db_medley.redis_conf import RedisConnector
from redis.exceptions import RedisError

from pypeline.barrier import LockingParallelBarrier
from pypeline.constants import DEFAULT_RESULT_TTL
from pypeline.dramatiq import REDIS_URL

from dramatiq.message import Message


class parallel_pipeline:
    """Chain actors together, passing the result of one actor to the
    next one in line.

    Parameters:
      children(typing.List[typing.List[Message]]): A sequence of messages or
        pipelines.  Child pipelines are flattened into the resulting
        pipeline.
      broker(Broker): The broker to run the pipeline on.  Defaults to
        the current global broker.
    """

    messages: list[Message]

    def __init__(self, messages: typing.List[typing.List[Message]], broker=None):
        self.broker = broker or get_broker()
        self.messages = messages
        self.execution_id = str(uuid4())
        execution_graph = []

        for message_group in self.messages:
            sub_execution_group = []
            group_completion_uuid = str(uuid4())
            for m in message_group:
                m.kwargs["event"]["execution_id"] = self.execution_id
                m.options["group_completion_uuid"] = group_completion_uuid
                message_dict = copy.deepcopy(m.asdict())
                sub_execution_group.append(message_dict)
            # Last item in the group is the id of the group to be executed
            execution_graph.append(sub_execution_group)

        self.execution_graph = execution_graph

        for m in self.messages[0]:
            m.options["execution_graph"] = execution_graph

    def __len__(self):
        """Returns the length of the parallel_pipeline."""
        count = 0
        for message_group in self.messages:
            count = count + len(message_group)

        return count

    def __str__(self):  # pragma: no cover
        """Return a string representation of the parallel_pipeline.

        This representation shows the order of execution for each group of messages.
        """
        result = []

        for i, message_group in enumerate(self.messages):
            group_str = f"Group {i + 1}: [\n"
            for j, message in enumerate(message_group):
                message_str = f"  Message {j + 1}: {message.actor_name}\n"
                group_str += message_str
            group_str += "]\n"
            result.append(group_str)

        return "".join(result)

    @property
    def completed(self):
        return self.completed_count == len(self)

    @property
    def completed_count(self):
        count = 0

        for message_group in self.messages:
            for message in message_group:
                try:
                    message.get_result()
                    count = count + 1
                except ResultMissing:
                    pass
        return count

    def run(self, *, delay=None):
        """Run this parallel_pipeline.

        Parameters:
          delay(int): The minimum amount of time, in milliseconds, the
            parallel_pipeline should be delayed by. If both parallel_pipeline's delay and
            first message's delay are provided, the bigger value will be
            used.

        Returns:
          parallel_pipeline: Itself.
        """
        starting_group = self.messages[0]

        completion_uuid = starting_group[0].options["group_completion_uuid"]
        locking_parallel_barrier = LockingParallelBarrier(
            REDIS_URL, task_key=completion_uuid, lock_key=f"{completion_uuid}-lock"
        )
        locking_parallel_barrier.set_task_count(len(starting_group))

        for m in starting_group:
            self.broker.enqueue(m, delay=delay)

        return self

    def get_result(self, *, block=False, timeout=None):
        """Get the result of this pipeline.

        Pipeline results are represented by the result of the last
        message in the chain.

        Parameters:
          block(bool): Whether or not to block until a result is set.
          timeout(int): The maximum amount of time, in ms, to wait for
            a result when block is True.  Defaults to 10 seconds.

        Raises:
          ResultMissing: When block is False and the result isn't set.
          ResultTimeout: When waiting for a result times out.

        Returns:
          object: The result.
        """
        last_message = self.messages[-1][-1]

        backend = self.broker.get_results_backend()
        return last_message.get_result(backend=backend, block=block, timeout=timeout)

    def get_results(self, *, block=False, timeout=None):
        """Get the results of each job in the pipeline.

        Parameters:
          block(bool): Whether or not to block until a result is set.
          timeout(int): The maximum amount of time, in ms, to wait for
            a result when block is True.  Defaults to 10 seconds.

        Raises:
          ResultMissing: When block is False and the result isn't set.
          ResultTimeout: When waiting for a result times out.

        Returns:
          A result generator.
        """
        deadline = None
        if timeout:
            deadline = time.monotonic() + timeout / 1000

        for message_group in self.messages:
            for message in message_group:
                if deadline:
                    timeout = max(0, int((deadline - time.monotonic()) * 1000))

                backend = self.broker.get_results_backend()
                yield {
                    message.actor_name: message.get_result(
                        backend=backend, block=block, timeout=timeout
                    )
                }

    def to_json(self) -> str:
        """Convert the execution graph to a JSON string representation.

        This method serializes the execution graph of the pipeline into a JSON string.
        This serialized form can be used to save the pipeline state or share it across different systems,
        enabling the retrieval of a pipeline "run" for obtaining its results at a later time.

        :return: A JSON string representing the execution graph.
        :rtype: str
        """
        return json.dumps(self.execution_graph)

    @classmethod
    def from_json(cls, json_data: str) -> parallel_pipeline:
        """Create a ParallelPipeline object from a JSON string representation of the execution graph.

        This class method deserializes a JSON string into a list of messages, each representing
        a task or operation in the pipeline. The method reconstructs the execution graph using
        the `dramatiq.message.Message` objects and returns an instance of the `parallel_pipeline` class.

        :param json_data: A JSON string containing the serialized execution graph.
        :type json_data: str
        :return: An instance of `parallel_pipeline` reconstructed from the JSON data.
        :rtype: parallel_pipeline
        """
        execution_graph = json.loads(json_data)

        messages = []

        for message_group in execution_graph:
            temp_group = []
            for message in message_group:
                temp_group.append(Message(**message))
            messages.append(temp_group)

        return cls(messages)


class PipelineResult:
    """
    A class to manage and retrieve the results of a parallel pipeline execution.

    The `PipelineResult` class provides methods for creating a result entry in a Redis database,
    loading pipeline data from Redis, and retrieving the status and results of the pipeline execution.

    Attributes:
        pipeline (parallel_pipeline): The pipeline object representing the execution graph.
        execution_id (str): A unique identifier for the execution of the pipeline.
        redis_key (str): The key used to store and retrieve pipeline data from Redis.
        redis_conn: A Redis connection object used to interact with the Redis database.
        result_ttl (int): Time-to-live (TTL) for the result entry in Redis, in seconds.
    """

    def __init__(self, execution_id: str, result_ttl: int = DEFAULT_RESULT_TTL):
        """
        Initialize a PipelineResult object with an execution ID and optional result TTL.

        :param execution_id: A unique identifier for the pipeline execution.
        :type execution_id: str
        :param result_ttl: The time-to-live (TTL) for the result entry in Redis. Defaults to DEFAULT_RESULT_TTL.
        :type result_ttl: int
        """
        self.pipeline: parallel_pipeline = None
        self.execution_id = execution_id
        self.redis_key = f"{execution_id}-results-key"
        self.redis_conn = RedisConnector().get_connection()
        self.result_ttl = result_ttl

    def create_result_entry(self, pipeline_json_str: str):
        """
        Store the serialized pipeline data in Redis with a specified TTL.

        This method saves the JSON string representation of the pipeline in the Redis database
        using the execution ID as the key. The entry is stored with a time-to-live (TTL) defined by `result_ttl`.

        :param pipeline_json_str: A JSON string representing the pipeline execution graph.
        :type pipeline_json_str: str
        :raises ValueError: If the provided pipeline data is None or an empty string.
        :raises RedisError: If there is an issue connecting to Redis or setting the value.
        """
        if not pipeline_json_str:
            raise ValueError("No pipeline data passed to create result store")

        try:
            self.redis_conn.setex(self.redis_key, self.result_ttl, pipeline_json_str)
        except RedisError as e:
            raise RuntimeError(f"Failed to store pipeline data in Redis: {e}")

    def load(self):
        """
        Load the pipeline data from Redis and reconstruct the pipeline object.

        This method retrieves the JSON string stored in Redis and deserializes it
        into a `parallel_pipeline` object, enabling access to the pipeline's execution details.

        :raises RedisError: If there is an issue connecting to Redis or retrieving the data.
        """
        try:
            pipeline_data = self.redis_conn.get(self.redis_key)
            if pipeline_data:
                self.pipeline = parallel_pipeline.from_json(pipeline_data)
            else:
                self.pipeline = None
        except RedisError as e:
            raise RuntimeError(f"Failed to load pipeline data from Redis: {e}")

    @property
    def status(self) -> str:
        """
        Get the current status of the pipeline execution.

        This property checks the completion status of the pipeline and returns its current state.

        :return: The status of the pipeline execution, which can be "complete", "pending", or "unavailable".
        :rtype: str
        """
        if not self.pipeline:
            return "unavailable"
        return "complete" if self.pipeline.completed else "pending"

    def get_results(self) -> dict:
        """
        Retrieve all results from the pipeline execution with unique actor identifiers.

        This method aggregates results from the pipeline and ensures that each actor's result
        has a unique identifier by appending a numeric suffix to duplicate actor names.

        :return: A dictionary containing all results from the pipeline execution, keyed by unique actor identifiers.
        :rtype: dict
        """
        if not self.pipeline:
            return {}

        results = {}
        for result in self.pipeline.get_results():
            for actor, res in result.items():
                unique_actor = self._get_unique_actor_name(actor, results)
                results[unique_actor] = res
        return results

    def get_result(self):
        """
        Retrieve a single result from the pipeline execution.

        This method returns the result of a single execution step from the pipeline, if available.

        :return: The result of a single execution step from the pipeline, or None if no pipeline is loaded.
        """
        if self.pipeline:
            return self.pipeline.get_result()

    def _get_unique_actor_name(self, actor: str, results: dict) -> str:
        """
        Generate a unique actor name by appending a numeric suffix if necessary.

        :param actor: The base name of the actor.
        :type actor: str
        :param results: The current dictionary of results to check for uniqueness.
        :type results: dict
        :return: A unique actor name.
        :rtype: str
        """
        if actor not in results:
            return actor

        suffix = 0
        new_actor = f"{actor}-{suffix}"
        while new_actor in results:
            suffix += 1
            new_actor = f"{actor}-{suffix}"
        return new_actor



""" Pypeline implementation as a Flask extension
"""
import os

if os.getenv("USE_GEVENT", "false").lower() == "true":
    import gevent.monkey

    gevent.monkey.patch_all()

import logging
from flask import Flask
from werkzeug.middleware.proxy_fix import ProxyFix
from flask_smorest import Api, Blueprint
from pypeline.extensions import pypeline_config
from pypeline.constants import DEFAULT_OPENAPI_CONFIG
from pypeline import __version__

logger = logging.getLogger(__name__)


class FlaskPypeline:
    """Pypeline Flask extension."""

    def __init__(self, app: Flask = None):
        """Class init"""
        self.app = app
        self.pypeline_config = pypeline_config if pypeline_config is not None else {}

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask, init_api: bool = False):
        """Pypeline bootstrapping process.

        Application config variables to set include:

            pypeline_client_version (default: v?.?.?)

        Optional, if `init_api` is True:

            API_DOCUMENTATION_TITLE
            API_DOCUMENTATION_DESCRIPTION
            OPENAPI_VERSION
            OPENAPI_URL_PREFIX
            OPENAPI_SWAGGER_APP_NAME
            OPENAPI_SWAGGER_UI_PATH
            OPENAPI_SWAGGER_BASE_TEMPLATE
            OPENAPI_SWAGGER_URL
            OPENAPI_SWAGGER_UI_URL
            SWAGGER_UI_DOC_EXPANSION
            EXPLAIN_TEMPLATE_LOADING

        Args:
            app (Flask): Flask Application to initialize.
            init_api (bool): If `True`, Pypeline will initialize its
                core APIs (including Pipelines, Scheduled Tasks, etc.) and
                provide a pre-configured OpenAPI Spec/Swagger UI interface
                available at the route defined in your application's config
                under `OPENAPI_URL_PREFIX` (default `/api`). Refer to
                [flask-smorest](https://flask-smorest.readthedocs.io/en/latest/openapi.html)
                documentation for additional configuration options.
        """
        # Ensure there's a PYPELINE_CLIENT_VERSION on app config
        app.config.setdefault(
            "PYPELINE_CLIENT_VERSION",
            app.config.get("PYPELINE_CLIENT_VERSION", "v?.?.?"),
        )

        app.wsgi_app = ProxyFix(app.wsgi_app)
        app.url_map.strict_slashes = False

        # Create and register the pypeline blueprint
        bp = Blueprint(
            "pypeline",
            __name__,
            template_folder="../templates",
            static_folder="../static",
            url_prefix="/pypeline",
        )
        app.register_blueprint(bp)

        # Bootstrap api if app requests
        if init_api is True:
            self._bootstrap_api(app)

    def _bootstrap_api(self, app: Flask):
        """If initializing the API, we will create the core Pypeline API paths
        and initialize the default Swagger documentation.
        """
        # Set sensible defaults for Swagger docs. Provided `app` will
        # take precedent.
        for swagger_config in DEFAULT_OPENAPI_CONFIG:
            app.config.setdefault(
                swagger_config[0], app.config.get(swagger_config[0], swagger_config[1])
            )

        # Attempt to override with values from client's pypeline.yaml if
        # they are available. This will add new tags and new docs if
        # defined and add to the core Pypeline API docs.
        api_config = self.pypeline_config.get("apiConfig", {})
        api_docs = api_config.get("apiDocumentation", {})

        custom_tags = api_config.get("prefixDescriptions", [])

        app.config["PYPELINE_CLIENT_VERSION"] = (
            api_docs.get("version", None)
            if api_docs.get("version", None) is not None
            else app.config["PYPELINE_CLIENT_VERSION"]
        )

        app.config["API_DOCUMENTATION_TITLE"] = (
            api_docs.get("title", None)
            if api_docs.get("title", None) is not None
            else app.config["API_DOCUMENTATION_TITLE"]
        )

        app.config["API_DOCUMENTATION_DESCRIPTION"] = (
            api_docs.get("description", None)
            if api_docs.get("description", None) is not None
            else app.config["API_DOCUMENTATION_DESCRIPTION"]
        )

        tags = [
            {"name": "Pipelines", "description": "Operations related to Pipelines"},
            {"name": "Schedules", "description": "Operations related to Schedules"},
        ] + custom_tags

        # Set up  the initializing spec kwargs for API
        spec_kwargs = {
            "title": app.config["API_DOCUMENTATION_TITLE"],
            "version": f"Pypeline: {__version__} - "
            f"Client: {app.config['PYPELINE_CLIENT_VERSION']}",
            "description": app.config["API_DOCUMENTATION_DESCRIPTION"],
            "tags": tags,
        }
        try:
            api = Api()
            api.init_app(app, spec_kwargs=spec_kwargs)

            self._register_api_namespaces(api)

        except Exception as e:
            api = None
            logging.exception(f"Unable to initialize API ... {e}")

        app.extensions.setdefault("pypeline_core_api", api)

    @staticmethod
    def _register_api_namespaces(api: Api):
        """Register Default API namespaces"""
        from pypeline.flask.api.pipelines import bp as pipelinesbp

        api.register_blueprint(pipelinesbp)
        from pypeline.flask.api.schedules import bp as schedulesbp

        api.register_blueprint(schedulesbp)
